/*
 * Copyright (C) 2015 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

const miss = require("mississippi");

import { Source, ISourceOptions } from "./source";
import { Transformer, ITransformerOptions } from "./transformer";
// import { Aggregate, IAggregateOptions } from "./aggregate";
import { Storage } from "./storage";

import { MongoStorage } from "../storage/mongo";
import { RedisStorage } from "../storage/redis";
import { MonetStorage } from "../storage/monet";

export interface IQuery {
    metrics: string[];
    dimensions: string[];
}

export class Server {
    public storage: {
        l1: Storage,
        l2: Storage,
        l3: Storage
    };
    public sources: Map<string, Source>;
    public transformers: Map<string, Transformer>;
    // public pipes: Array<[Source, Transformer]>;
    // public aggregates: Map<string, Aggregate>;

    constructor() {
        this.storage = {
            l1: new MongoStorage(),
            l2: new RedisStorage(),
            l3: new MonetStorage()
        };

        this.sources = new Map();
        this.transformers = new Map();
        // this.pipes = new Array();
        // this.aggregates = new Map();
    }

    public source(name: string, options?: ISourceOptions): Source {
        if (typeof options !== "undefined") {
            if (this.sources.has(name)) {
                throw new Error("A source named '" + name + "' already exists");
            }

            const source = new Source(this, options);
            this.sources.set(name, source);
            return source;
        }
        else {
            if (!this.sources.has(name)) {
                throw new Error("A source named '" + name + "' does not exist");
            }

            return this.sources.get(name);
        }
    }

    public transformer(name: string, options?: ITransformerOptions): Transformer {
        if (typeof options !== "undefined") {
            if (this.transformers.has(name)) {
                throw new Error("A transformer named '" + name + "' already exists");
            }

            const transformer = new Transformer(this, options);
            this.transformers.set(name, transformer);
            return transformer;
        }
        else {
            if (!this.transformers.has(name)) {
                throw new Error("A transformer named '" + name + "' does not exist");
            }

            return this.transformers.get(name);
        }
    }

    // public pipe(source: Source, transformer: Transformer): void {
    //     function findPipe(pipe: [Source, Transformer]) {
    //         return (pipe[0] === source) || (pipe[1] === transformer);
    //     }
    //
    //     if (this.pipes.find(findPipe)) {
    //         throw new Error("A pipe containing the same source or destination already exists");
    //     }
    //
    //     this.pipes.push([source, transformer]);
    // }

    public query(query: IQuery, callback: Function) {
        let coverageMap = this.buildCoverageMap(query);

        // check if all the metrics are covered by the existing transfomers
        for (const metric of query.metrics) {
            if (!coverageMap.has(metric)) {
                callback("The metric " + metric + " is not covered by any transformer.");
                return;
            }
        }

        // list all the transformers we are going to need
        let transformers = new Map();
        for (const [, t] of coverageMap) {
            if (!transformers.has(t[0].id)) {
                transformers.set(t[0].id, t[0]);
            }
        }

        // apply transformers
        for (const transformer of transformers.values()) {
            transformer.query(query.metrics, query.dimensions);
            // function write(data: any, enc: string, cb: Function) {
            //     console.log(data);
            //     cb();
            // }
            //
            // function flush(cb: Function) {
            //     cb();
            // }
            //
            // miss.pipeline(
            //     transformer.source.stream(),
            //     transformer.position(),
            //     transformer.measure(query.metrics),
            //     transformer.aggregate(query.dimensions),
            //     miss.to.obj(write, flush)
            // );
        }

        callback(null, []);
    }

    private buildCoverageMap(query: IQuery): Map<string, Transformer[]> {
        let coverageMap: Map<string, Transformer[]> = new Map();

        // checks if l1 is covered by l2
        function isCovered(l1: string[], l2: string[]): boolean {
            for (const s1 of l1) {
                let res = l2.find((s2: string) => {
                    return s2 === s1;
                });
                if (typeof res === "undefined") {
                    return false;
                }
            }
            return true;
        }

        // find the transformers that cover the query
        for (const [, transformer] of this.transformers) {
            if (isCovered(query.dimensions, transformer.dimensions)) {
                for (const metric of transformer.metrics) {
                    if (query.metrics.indexOf(metric) < 0) {
                        // if the metric is not queried, it doesn't need to be
                        // added to the coverage map
                        continue;
                    }

                    if (!coverageMap.has(metric)) {
                        coverageMap.set(metric, [transformer]);
                    }
                    else {
                        coverageMap.get(metric).push(transformer);
                    }
                }
            }
        }

        return this.sortCoverageMap(query, coverageMap);
    }

    private sortCoverageMap(query: IQuery, coverageMap: Map<string, Transformer[]>): Map<string, Transformer[]> {
        let newCoverageMap: Map<string, Transformer[]> = new Map();

        let numAppearences: { [tId: string]: number; } = {};

        // attribute a point for every appearence of a transformer in the map
        for (const [, transformers] of coverageMap) {
            for (const transformer of transformers) {
                numAppearences[transformer.id]++;
            }
        }

        // select the best transformer for each metric
        for (const [metric, transformers] of coverageMap) {
            let newTransformers = transformers.sort((a, b) => {
                let aAppear = numAppearences[a.id] / query.metrics.length;
                let bAppear = numAppearences[b.id] / query.metrics.length;
                let aDist = Math.max((a.dimensions.length - query.dimensions.length) * 0.1, 1);
                let bDist = Math.max((b.dimensions.length - query.dimensions.length) * 0.1, 1);

                let aPoints = aAppear * 0.5 + (1 - aDist) * 0.5;
                let bPoints = bAppear * 0.5 + (1 - bDist) * 0.5;

                return (aPoints > bPoints) ? -1 : 1;
            });

            newCoverageMap.set(metric, newTransformers);
        }

        return newCoverageMap;
    }
}
