/*
 * Copyright (C) 2015 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

const miss = require("mississippi");

import { expect } from "chai";
import { Hash } from "../util/hash";
import { Server } from "./server";
import { Transformer } from "./transformer";

describe("transformer class", () => {
    const mockServer = new (class MockServer extends Server {
        //
    })();

    const transformer = new Transformer(mockServer, {
        dimensions: ["dim:one", "dim:two"],
        metrics: ["met:one"],
        operators: {
            position: ((data: any) => {
                return {
                    "dim:one": data.id,
                    "dim:two": Hash.sha1(data.seed)
                };
            }),
            measure: ((data: any) => {
                return {
                    "met:one": Math.floor(data.seed * 100000) / 100
                };
            }),
            aggregate: {
                "met:one": "sum"
            }
        }
    });

    it("should be able to retrieve dimensions with by using a position operator", (callback: Function) => {
        let i = 0;
        let source = miss.from.obj((size: Number, next: Function) => {
            if (i >= 1024) {
                next(null, null);
                return;
            }

            next(null, {
                doc: {
                    id: i % 10,
                    seed: ((i % 10) === 0) ? 5 : Math.random()
                }
            });
            i++;
        });

        let sink = miss.to.obj((data: any, enc: string, cb: Function) => {
            expect(data).to.have.property("dimensions");
            expect(data).to.have.property("doc");
            expect(data.dimensions).to.have.property("dim:one");
            expect(data.dimensions).to.have.property("dim:two");
            cb();
        }, (err: any) => {
            callback();
        });

        miss.pipeline(source, transformer.position(), sink);
    });

    it("should be able to retrieve metrics with by using a measure operator", (callback: Function) => {
        let i = 0;
        let source = miss.from.obj((size: Number, next: Function) => {
            if (i >= 1024) {
                next(null, null);
                return;
            }

            next(null, {
                dimensions: {
                    "dim:one": i % 10,
                    "dim:two": Hash.sha1(((i % 10) === 0) ? 5 : Math.random())
                },
                doc: {
                    id: i % 10,
                    seed: ((i % 10) === 0) ? 5 : Math.random()
                }
            });
            i++;
        });

        let sink = miss.to.obj((data: any, enc: string, cb: Function) => {
            expect(data).to.have.property("dimensions");
            expect(data).to.have.property("metrics");
            expect(data.metrics).to.have.property("met:one");
            cb();
        }, (err: any) => {
            callback();
        });

        miss.pipeline(source, transformer.measure(), sink);
    });
});
