/*
 * Copyright (C) 2015 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

const miss = require("mississippi");
const uuid = require("node-uuid");

import { Server } from "./server";
import { Source } from "./source";

export interface ITransformerOptions {
    dimensions: string[];
    metrics: string[];
    operators: {
        position: (doc: any) => any;
        measure: (doc: any) => any;
        aggregate: (doc: any, newDimensions: string[]) => any;
    };
}

export class Transformer {
    public id: string;
    public dimensions: string[];
    public metrics: string[];

    private server: Server;
    private operators: any;
    private source: Source;

    constructor(server: Server, options: ITransformerOptions) {
        this.id = uuid.v4();
        this.server = server;
        this.dimensions = options.dimensions;
        this.metrics = options.metrics;
        this.operators = options.operators;
        this.source = null;
    }

    public setSource(source: Source) {
        this.source = source;
    }

    public position() {
        if (!this.source) {
            throw new Error("Source not defined");
        }

        return miss.through.obj((data: any, enc: string, callback: Function) => {
            let dimensions = this.operators.position(data.doc);
            callback(null, { dimensions: dimensions, doc: data.doc });
        });
    }

    public measure() {
        if (!this.source) {
            throw new Error("Source not defined");
        }

        return miss.through.obj((data: any, enc: string, callback: Function) => {
            let metrics = this.operators.measure(data.doc);
            callback(null, { dimensions: data.dimensions, metrics: metrics });
        });
    }

    public aggregate(newDimensions: string[]) {
        if (!this.source) {
            throw new Error("Source not defined");
        }

        return miss.through.obj((data: any, enc: string, callback: Function) => {
            let newMetrics = this.operators.aggregate(data, newDimensions);
            callback(null, { dimensions: newDimensions, metrics: newMetrics });
        });
    }

    public query(queryMetrics: string[], queryDimensions: string[]) {
        if (!this.baseView) {
            this.baseView = new View(this.metrics, this.dimensions);

            miss.pipeline(
                this.source.stream(),
                this.position(),
                this.measure(),
                this.baseView.sink()
            );
        }

        this.baseView.aggregate(queryMetrics, queryDimensions);
    }
}
